/*=======================================================
  Filename: CangoAxes-1v03.js
  Rev 1
  By: A.R.Collins

  Description: Adds Axes drawing methods and Arrow
  drawing methods and drawHTMLtext to the Cango graphics
  library. Allso adds the global utility functions:
  sprintf, toEngFixed, toEngPrec, engNotation.

  License: Released into the public domain
  latest version at
  <http://www/arc.id.au/>

  Date    Description                               |By
  ------------------------------------------------------
  30Apr14 First beta                                 ARC
  06May14 Don't set color=null to get cgo.penCol     ARC
  10May14 Added sprintf
          Added arrowArc                             ARC
  11May14 Rev 1.00 released                          ARC
  30May14 Mod to handle no fontWeight in compileText ARC
  25Jul14 Use Javascript Module format               ARC
  29Jul14 Abandon isoGC it gives rounding errors     ARC
 =======================================================*/

var sprintf, toEngFixed, toEngPrec, AxisTicsAuto, AxisTicsManual;

Cango = (function(CangoCore)
{
  "use strict";

  var isArray = function(obj)
  {
    return Object.prototype.toString.call(obj) === '[object Array]';
  };

  var isNumber = function(o)
  {
    return !isNaN(o) && o !== null && o !== "" && o !== false;
  };

  function toEngFixed(val, decPlaces)      // rounds to X dec places and no stripping of 0s
  {
    var unit = "pnum kMG",
        man, pwr,
        expt = 0,
        str = "";

    if ((decPlaces === undefined)||(decPlaces<0)||(decPlaces>10))
    {
      decPlaces = 2;
    }
    man = 0.0;
    if (Math.abs(val)>1.0E-12)
    {
      pwr = Math.log(Math.abs(val))/(3.0*Math.LN10);
      expt = Math.floor(pwr);
      man = val/Math.pow(1000.0, expt);
      expt *= 3;
    }
    // now force round to decPlaces
    str = man.toFixed(decPlaces);
    // now add the symbol for the exponent
    return str+unit.charAt(expt/3+4);
  }

  toEngPrec = function(val, sigFigs)      // rounds to X significant figures and no stripping of 0s
  {
    var unit = "pnum kMG",
        man = 0,
        pwr,
        expt = 0,
        str = "";

    if (Math.abs(val)>1.0E-12)
    {
      pwr = Math.log(Math.abs(val))/(3.0*Math.LN10);
      expt = Math.floor(pwr);
      man = val/Math.pow(1000.0, expt);
      expt *= 3;
    }
    str = man.toPrecision(sigFigs);
    // now add the symbol for the exponent
    return str+unit.charAt(expt/3+4);
  };

  var engNotation = function(val, tenthsOK)        // rounds to 2 dec places and strips trailing 0s
  {
    var unit = "pnum kMG",
    man = 0.0,
    pwr,
    expt = 0,
    str = "";

    if (Math.abs(val)>1.0E-12)
    {
      if (tenthsOK)
      {
        pwr = Math.log(Math.abs(10*val))/(3.0*Math.LN10); // calc exp on 10 x val allows .9 not 900m
      }
      else
      {
        pwr = Math.log(Math.abs(val))/(3.0*Math.LN10);
      }
      expt = Math.floor(pwr);
      man = val/Math.pow(1000.0, expt);
      expt *= 3;
    }
    // now force round to decPlaces
    str = man.toFixed(2);
    // now strip trailing 0s
    while (str.charAt(str.length-1) === '0')
    {
      str = str.substring(0,str.length-1);
    }
    if (str.charAt(str.length-1) === '.')
    {
      str = str.substring(0,str.length-1);
    }
    // now add the symbol for the exponent
    if (expt)
    {
      return str+unit.charAt(expt/3+4);
    }

    return str;                   // dont add unnecessary space
  };

  AxisTicsAuto = function(mn, mx)   // optional 'numTics' forces the value of numSteps
  {
    /* Calculate the tic mark spacing for graph plotting
     * Given the minimum and maximum values of the axis
     * returns the first tic value and the tic spacing.
     * The algorithm gives tic spacing of 1, 2, 5, 10, 20 etc
     * and a number of ticks from ~5 to ~11
     */
    var pwr, spanman, stepman,
        spanexp, stepexp;

    this.tic1 = 0;
    this.ticStep = 0;

    if (mn>=mx)
    {
      alert("Max must be greater than Min!");
      return;
    }

    pwr = Math.log(mx-mn)/Math.LN10;
    if (pwr<0.0)
    {
      spanexp = Math.floor(pwr) - 1;
    }
    else
    {
      spanexp = Math.floor(pwr);
    }
    spanman = (mx-mn)/Math.pow(10.0, spanexp);
    if(spanman>=5.5)
    {
      spanexp += 1;
      spanman /= 10.0;
    }
    stepman = 0.5;
    if(spanman<2.2)
    {
      stepman = 0.2;
    }
    if(spanman<1.1)
    {
      stepman = 0.1;
    }
    stepexp = 3*Math.floor(spanexp/3);
    if((spanexp < 0)&&(spanexp%3 !== 0))
    {
      stepexp -= 3;
    }
    stepman = stepman*Math.pow(10.0, (spanexp-stepexp));
    this.ticStep = stepman*Math.pow(10.0, stepexp);
    if(mn>=0.0)
    {
      this.tic1 = (Math.floor((mn/this.ticStep)-0.01)+1)*this.ticStep;   // avoid math noise
    }
    else
    {
      this.tic1 = -Math.floor((-mn/this.ticStep)+0.01)*this.ticStep;   // avoid math noise
    }

  /*  var str = "";
      str += "ymin="+mn+"  ymax="+mx+"\n";
      str += "tic1= "+this.tic1+ "\n";
      str += "ticStep= "+this.ticStep+ "\n";
      alert(str);
  */
  };

  AxisTicsManual = function(xmin, xmax,	xMn, xMj)
	{
		var dx;		// tolerance for avoiding maths noise

    this.tic1 = 0;
    this.ticStep = 0;
    this.lbl1 = 0;
    this.lblStep = 0;
    this.ticScl = 0;     // reserved for future use

		if ((xmin===undefined)||(xmax===undefined)||(xMn===undefined)||(xMn<=0))
    {
			return;
    }

		dx = 0.01*xMn;
		this.tic1 = xMn*Math.ceil((xmin-dx)/xMn);
    this.ticStep = xMn;

		if ((xMj!==undefined)&&(xMj>0))
		{
			this.lblStep = this.ticStep*Math.round(xMj/xMn);
			dx = 0.01*xMn;
			this.lbl1 = this.lblStep*Math.ceil((xmin-dx)/this.lblStep);
		}
    // OPTION:
    // to make all labels have same scale factor, calc lastTic and corresponding tag "m kMG" etc
    // calc engnotation for xTic1 exp=xTicScl, tag=xTicTag
    // plot x = xtic1 + n*xTicStep
    // label x/xTicScl+xTicTag
	};

  CangoCore.prototype.drawAxes = function(xOrg, yOrg, xMin, xMax, yMin, yMax, xMinTic, yMinTic, xMajTic, yMajTic, x10ths, y10ths)
  {
    var x, y,
        pos,
        data = [],
        ticLen = 0.007*this.vpW/this.xscl,     // 0.7% of width of canvas
        xLblOfs = -0.025*this.vpW/this.yscl,   // 2.5% of width of canvas
        yLblOfs = 0.025*this.vpW/this.xscl,    // 2.5% of width of canvas
      	lorg = 1,
        side = 1,
        tickCmds, bigTickCmds,
        xTics, yTics;

    // draw all ticks in world coords set using this.toViewportCoords().

    tickCmds = this.compilePath(['M', -ticLen, 0, 'L', ticLen, 0], null, 1, 'iso');
    bigTickCmds = this.compilePath(['M', -2*ticLen, 0, 'L', 2*ticLen, 0], null, 1, 'iso');
    xTics = new AxisTicsManual(xMin, xMax, xMinTic, xMajTic);
    yTics = new AxisTicsManual(yMin, yMax, yMinTic, yMajTic);

    // draw axes
    data = ['M', xMin, yOrg, 'L', xMax, yOrg, 'M', xOrg, yMin, 'L', xOrg, yMax];
    this.drawPath(data);
    // X axis tick marks
    if (xTics.ticStep)
    {
      for(x=xTics.tic1; x<=xMax; x+=xTics.ticStep)
      {
        this.render(tickCmds, x, yOrg, 1, 90);     // maintain aspect ratio, rotate 90deg
      }
    }
    // Y axis tick marks
    if (yTics.ticStep)
    {
      for(y=yTics.tic1; y<=yMax; y+=yTics.ticStep)
      {
        this.render(tickCmds, xOrg, y, 1, 0);
      }
    }
    // major ticks X axis
    if (xTics.lblStep)
    {
      for(x=xTics.lbl1; x<=xMax; x+=xTics.lblStep)
      {
        this.render(bigTickCmds, x, yOrg, 1, 90);
      }
    }
    // major ticks Y axis
    if (yTics.lblStep)
    {
      for(y=yTics.lbl1; y<=yMax; y+=yTics.lblStep)
      {
        this.render(bigTickCmds, xOrg, y, 1, 0);
      }
    }
    // now label the axes
		if (xTics.lblStep)
		{
    	// X axis, decide whether to label above or below X axis
  		if (((yOrg<yMin+0.55*(yMax-yMin)) && (this.yscl<0))||((yOrg>yMin+0.45*(yMax-yMin)) && !(this.yscl<0)))
    	{   // x axis on bottom half of screen
    		side = -1;
    		lorg = 2;
    	}
    	else
    	{
    	  side = 1;
    		lorg = 8;
    	}
    	for (x = xTics.lbl1; x<=xMax; x += xTics.lblStep)
    	{
  			// skip label at the origin if it would be on the other axis
  			if ((x===xOrg)&&(yOrg>yMin)&&(yOrg<yMax))
        {
  				continue;
        }
        this.drawText(engNotation(x, x10ths), x, yOrg+side*xLblOfs, null, 10, lorg);  // use default strokeColor
    	}
    }

		if (yTics.lblStep)
		{
    	// Y axis, decide whether to label to right or left of Y axis
    	if (((xOrg<xMin+0.5*(xMax-xMin)) && (this.xscl>0))||((xOrg>xMin+0.5*(xMax-xMin)) && !(this.xscl>0)))
    	{  // y axis on left half of screen
    		side = -1;
    		lorg = 6;
    	}
    	else
    	{
    	  side = 1;
    		lorg = 4;
    	}
    	for (y = yTics.lbl1; y<=yMax; y += yTics.lblStep)
    	{
  			// skip label at the origin if it would be on the other axis
  			if ((y===yOrg)&&(xOrg>xMin)&&(xOrg<xMax))
        {
  				continue;
        }
        this.drawText(engNotation(y, y10ths), xOrg+side*yLblOfs, y, null, 10, lorg);
    	}
    }
  };

  CangoCore.prototype.drawBoxAxes = function(xMin, xMax, yMin, yMax, xMn, yMn, xUnits, yUnits, title)
  {
		var x, y,
        pos,
        data = [],
        ticLen = 0.01*this.vpW/this.xscl,      // 1.0% of width of canvas
        xLblOfs = -0.016*this.vpW/this.yscl,   // 1.6% of width of canvas
        yLblOfs = 0.016*this.vpW/this.xscl,    // 1.6% of width of canvas
      	lbl,
        tickCmds,
    		xTics, yTics;

    // draw all ticks in these world coords using this.toViewportCorrds().
    this.setPropertyDefault('strokeColor', '#cccccc');

    tickCmds = this.compilePath(['M', -ticLen, 0, 'L', 0, 0], 'rgba(255,255,255,1.0)', 1, 'iso');
		xTics = new AxisTicsManual(xMin, xMax, xMn);
		yTics = new AxisTicsManual(yMin, yMax, yMn);
		// Draw box axes
    data = ['M', xMin, yMin, 'L', xMin, yMax, xMax, yMax, xMax, yMin, 'z'];
    this.drawPath(data, 0, 0, '#cccccc');

		if (xMn === undefined)
    {
			return;
    }
  	for (x=xTics.tic1; x<=xMax; x += xTics.ticStep)
  	{
      this.render(tickCmds, x, yMin, 1, 90);      // just draw the tick mark
      if ((x !== xMin)&&(x !== xMax))        // no dots on the box
      {
        this.drawPath(['M', x, yMin, 'L', x, yMax], 0, 0, 'rgba(255,255,255,0.2)');
      }
  	}
		if (yMn === undefined)
    {
			return;
    }
  	for (y=yTics.tic1; y<=yMax; y += yTics.ticStep)
  	{
      this.render(tickCmds, xMin, y, 1, 0);      // just draw the tick mark
      if ((y !== yMin)&&(y !== yMax))
      {
        this.drawPath(['M', xMin, y, 'L', xMax, y], 0, 0, 'rgba(255,255,255,0.2)');
      }
  	}
		// Now labels, X axis, label only first and last tic below X axis
		x = xTics.tic1;
    this.drawText(engNotation(x), x, yMin - xLblOfs, '#cccccc', 10, 1);
  	while(x+xTics.ticStep <= 1.05*xMax)
    {
  		x += xTics.ticStep;
    }
    pos = this.toViewportCoords(x, yMin);
    this.drawText(engNotation(x), x, yMin - xLblOfs, '#cccccc', 10, 3);
		// Y axis, label bottom and top tics to left of Y axis
 		y = yTics.tic1;
    pos = this.toViewportCoords(xMin, y);
    this.drawText(engNotation(y), xMin - yLblOfs, y, '#cccccc', 10, 6);
  	while (y + yTics.ticStep <= 1.05*yMax)
    {
			y += yTics.ticStep;
    }
    this.drawText(engNotation(y), xMin - yLblOfs, y, '#cccccc', 10, 6);
    // x axis label
    if (typeof xUnits === "string")
    {
      lbl = engNotation(xTics.ticStep)+xUnits+"/div";
    }
    else
    {
      lbl = engNotation(xTics.ticStep)+"/div";
    }
    x = xMin+(xMax-xMin)/2;
    pos = this.toViewportCoords(x, yMin);
    this.drawText(lbl, x, yMin - xLblOfs, '#cccccc', 10, 2);
    // y axis label
    if (typeof yUnits === "string")
    {
      lbl = engNotation(yTics.ticStep)+yUnits;
    }
    else
    {
      lbl = engNotation(yTics.ticStep);
    }
    y = yMin+(yMax-yMin)/2;
    pos = this.toViewportCoords(xMin, y);
    this.drawText(lbl, xMin - yLblOfs, y, '#cccccc', 10, 9);
    y = yMin+(yMax-yMin)/2.15;
    this.drawText("/div", xMin - yLblOfs, y, '#cccccc', 10, 3);
    // title
    if (typeof title === "string")
    {
      this.drawText(title, xMin, yMax + xLblOfs, '#cccccc', 10, 7);
    }
	};

  CangoCore.prototype.drawXYAxes = function(xOrg, yOrg, xMin, xMax, yMin, yMax, xUnits, yUnits, xLabel, yLabel)
  {
    var pos,
        dataCmds,
        xOfs = 6, yOfs = 9,
      	lorg = 1,
        side = 1,      // 1 or -1 depending on the side of the axis to label
        xTics, yTics,
        xU = "", yU = "",
        xL = "", yL = "",
        isoGC;
    // to get uniform tick lengths and label positions create a isotropic set of world coords where x axis is 100 units
    isoGC = new Cango(this.cId);   // copy the canvas ID from current context
    isoGC.dupCtx(this);
    isoGC.setViewport();           // full screen
    isoGC.setWorldCoords(0, 0, 100); // viewport coords (x axis 100 units y axis same unit)
    // draw all ticks in world coords using this.toViewportCorrds().
    xTics = new AxisTicsAuto(xMin, xMax);
    yTics = new AxisTicsAuto(yMin, yMax);

    this.drawAxes(xOrg, yOrg, xMin, xMax, yMin, yMax, xTics.ticStep, yTics.ticStep, 2*xTics.ticStep, 2*yTics.ticStep);
    if (xUnits && (xUnits.length>0))
    {
      xU = "("+xUnits+")";
    }
    if (yUnits && (yUnits.length>0))
    {
      yU = "("+yUnits+")";
    }

    if ((xLabel !== undefined)&&(xLabel.length>0))
    {
      xL = xLabel;
    }
    if (((yOrg<yMin+0.55*(yMax-yMin)) && (this.yscl<0))||((yOrg>yMin+0.45*(yMax-yMin)) && !(this.yscl<0)))
    {
      side = -1;
      lorg = 3;
    }
    else
    {
      side = 1;
      lorg = 9;
    }
    pos = this.toViewportCoords(xMax, yOrg);
    dataCmds = isoGC.compileText(xL+xU, null, null, lorg);
    isoGC.render(dataCmds, pos.x, pos.y + side*xOfs);

    if (yLabel &&(yLabel.length>0))
    {
      yL = yLabel;
    }
  	// Y axis, decide whether to label to right or left of Y axis
  	if (((xOrg<xMin+0.5*(xMax-xMin)) && (this.xscl>0))||((xOrg>xMin+0.5*(xMax-xMin)) && !(this.xscl>0)))
  	{
  		// y axis on left half of screen
  		side = -1;
      lorg = (this.yscl<0)? 9: 7;
  	}
  	else
  	{
  	  side = 1;
      lorg = (this.yscl<0)? 3: 1;
  	}
    pos = this.toViewportCoords(xOrg, yMax);
    dataCmds = isoGC.compileText(yL+yU, null, null, lorg);
    isoGC.render(dataCmds, pos.x + side*yOfs, pos.y, 1, 90);
  };

  CangoCore.prototype.drawHTMLtext = function(str, x, y, color, fontSz, fontWt)
  {
    function createHTMLoverlay(gc)
    {
      var ovlHTML, newOvl,
          w = gc.rawWidth,
          h = gc.rawHeight,
          ovlId,
          cvs = gc.cnvs,
          currOvl;

      ovlId = gc.cId+"_aovl_";
      // create 1px square DIV place at top left to give position reference to HTML children
      ovlHTML = "<div id='"+ovlId+"' style='position:absolute; width:1px; height:1px;'></div>";
      if (document.getElementById(ovlId))
      {
        currOvl = document.getElementById(ovlId);
        currOvl.parentNode.removeChild(currOvl);
      }
      cvs.insertAdjacentHTML('afterend', ovlHTML);
      // make it the same size as the background canvas
      newOvl = document.getElementById(ovlId);
      newOvl.style.backgroundColor = 'transparent';
      newOvl.style.left = gc.cnvs.offsetLeft+'px';
      newOvl.style.top = gc.cnvs.offsetTop+'px';
      newOvl.style.fontFamily = gc.fontFamily;
      newOvl.style.lineHeight = '1.4em';

      return newOvl;
    }

    var size = fontSz || this.fontSize,   // fontSize in pts
        weight = this.fontWeight,         // default = 400
        topPx, leftPx,
        xOfs = x || 0,
        yOfs = y || 0,
        txtCol = this.penCol,
        divNode;

    if (document.getElementById(this.cId+"_aovl_") === null)
    {
      this.cnvs.alphaOvl = createHTMLoverlay(this);
    }
    if (typeof str !== 'string')
    {
      return;
    }
    if (typeof fontWt === 'string')
    {
      weight = fontWt;           // 'bold' etc
    }
    else if (isNumber(fontWt) && (fontWt >= 100) && (fontWt <= 900))
    {
      weight = fontWt;           // 100 .. 900
    }
    if (typeof color === "string")
    {
      txtCol = color;
    }

    size *= 1.3;    // size = 1em, convert to pixels

    divNode = document.createElement("div");
    divNode.style.position = "absolute";
    divNode.style.backgroundColor = "transparent";
    // to calc label top position: add viewport offsets + world coords*scl offset from vp lowerleft
    topPx = (this.vpLLy+yOfs*this.yscl+this.yoffset).toFixed(0);
    leftPx = (this.vpLLx+xOfs*this.xscl+this.xoffset).toFixed(0);
    // style the div depending of the lorg value eg set text-align to left right or center
    divNode.style.cssText += "top:"+topPx+"px; left:"+leftPx+"px; color:"+txtCol+"; font-size:"+size+"px; font-weight:"+weight;
    divNode.innerHTML = str;
    this.cnvs.alphaOvl.appendChild(divNode);
  };

  CangoCore.prototype.arrow = function(x1, y1, x2, y2, size) /* from x1,y1 to x2,y2 */
  {
    var a = 20*Math.PI/180.0,  // half included angle of arrow head = 20deg
        scale = size || 1.5,
        ds = (scale-1)/2,      // correct shaft width for outline
        dx = (x2-x1)*this.xscl,         // pixel coords
        dy = (y2-y1)*this.yscl,
        theta = -Math.atan2(dy, dx), // angle of the arrow to x axis
        phiL = theta + Math.PI + a,
        phiR = theta + Math.PI - a,
        phiC = theta + Math.PI,
        x3, x4, x5, x6, x7, x8,
        y3, y4, y5, y6, y7, y8,
        r, rs;

    if (scale<1)
    {
      scale = 1;
    }
    r = 4+2.75*scale
    rs = r*Math.cos(a);                   // length of arrow head along shaft

    x3 = x2 + r*Math.cos(phiL)/this.xscl;
    y3 = y2 - r*Math.sin(phiL)/this.yscl;
    x4 = x2 + r*Math.cos(phiR)/this.xscl;
    y4 = y2 - r*Math.sin(phiR)/this.yscl;
    x5 = x2 + (rs*Math.cos(phiC) - ds*Math.sin(phiC))/this.xscl;
    y5 = y2 - (rs*Math.sin(phiC) + ds*Math.cos(phiC))/this.yscl;
    x6 = x1 - ds*Math.sin(phiC)/this.xscl;
    y6 = y1 - ds*Math.cos(phiC)/this.yscl;
    x7 = x1 + ds*Math.sin(phiC)/this.xscl;
    y7 = y1 + ds*Math.cos(phiC)/this.yscl;
    x8 = x2 + (rs*Math.cos(phiC) + ds*Math.sin(phiC))/this.xscl;
    y8 = y2 - (rs*Math.sin(phiC) - ds*Math.cos(phiC))/this.yscl;

    return ["M", x2, y2, "L", x3, y3, x5, y5, x6, y6, x7, y7, x8, y8, x4, y4, x2, y2];
  };

  CangoCore.prototype.arrowArc = function(r, startA, stopA, size)
  {
    // will create an arc centred on (0,0) radius r, from angle startA to stopA (deg)
    // arrow head will be at r,stopA line width =size aroow head in proprtion
    var b1x, b1y, e1x, e1y,
        b2x, b2y, e2x, e2y,
        lrg, sweep,
        rad = Math.PI/180,
        scale = size || 1.5,
        ds = (scale-1)/2,      // correct shaft width for outline
        r1, r2, ln, lnRad,
        stopRad = rad*stopA,
        startRad = rad*startA,
        spanRad = rad*Math.abs(startA-stopA),
        baseA,
        tx, ty,               // tip x,y
        qr1, qr2,             // radii of tips of barbs
        q1x, q1y, q2x, q2y,   // tips of arrow barbs
        sgnY = (this.yscl<0)? 1: -1;

    r1 = r-ds/this.xscl;
    r2 = r+ds/this.xscl;
    ln = (3.5+2*scale)/this.xscl;     // length of arrow head along arc (now in world coords)
    lnRad = ln/r;                       // length of arrow head in radians
    if (lnRad>spanRad)
    {
      lnRad = spanRad;
    }
    baseA = (stopRad>startRad)? stopRad-lnRad: stopRad+lnRad;
    qr1 = r-(1.0+scale)/this.xscl;
    qr2 = r+(1.0+scale)/this.xscl;

    b1x = r1*Math.cos(startRad);
    b1y = r1*Math.sin(startRad)*sgnY;
    e1x = r1*Math.cos(baseA);
    e1y = r1*Math.sin(baseA)*sgnY;
    b2x = r2*Math.cos(startRad);
    b2y = r2*Math.sin(startRad)*sgnY;
    e2x = r2*Math.cos(baseA);
    e2y = r2*Math.sin(baseA)*sgnY;
    tx = r*Math.cos(stopRad);
    ty = r*Math.sin(stopRad)*sgnY;
    q1x = qr1*Math.cos(baseA);
    q1y = qr1*Math.sin(baseA)*sgnY;
    q2x = qr2*Math.cos(baseA);
    q2y = qr2*Math.sin(baseA)*sgnY;

    lrg = (Math.abs(stopA - startA) > 180)? 1: 0;
    if (this.yscl<0)
    {
      sweep = (startA > stopA) ? 1: 0;
    }
    else
    {
      sweep = (startA > stopA) ? 0: 1;
    }

    return ["M", tx,ty, "L", q1x,q1y, e1x,e1y, "A",r1,r1,0,lrg,1-sweep,b1x,b1y, "L", b2x,b2y, "A",r2,r2,0,lrg,sweep,e2x,e2y, "L", q2x,q2y, "A",r2,r2,0,0,sweep,tx,ty];
  }

      /* ==========================================================================
    // http://kevin.vanzonneveld.net
    // +   original by: Ash Searle (http://hexmen.com/blog/)
    // + namespaced by: Michael White (http://getsprink.com)
    // +    tweaked by: Jack
    // +   improved by: Kevin van Zonneveld (http://kevin.vanzonneveld.net)
    // +      input by: Paulo Freitas
    // +   improved by: Kevin van Zonneveld (http://kevin.vanzonneveld.net)
    // +      input by: Brett Zamir (http://brett-zamir.me)
    // +   improved by: Kevin van Zonneveld (http://kevin.vanzonneveld.net)
    // +   improved by: Dj
    // +   improved by: Allidylls
    // *     example 1: sprintf("%01.2f", 123.1);
    // *     returns 1: 123.10
    // *     example 2: sprintf("[%10s]", 'monkey');
    // *     returns 2: '[    monkey]'
    // *     example 3: sprintf("[%'#10s]", 'monkey');
    // *     returns 3: '[####monkey]'
    // *     example 4: sprintf("%d", 123456789012345);
    // *     returns 4: '123456789012345'
   *==========================================================================*/


  sprintf = function()
  {
    var regex = /%%|%(\d+\$)?([\-\+\'#0 ]*)(\*\d+\$|\*|\d+)?(\.(\*\d+\$|\*|\d+))?([scboxXuideEfFgG])/g,
        a = arguments,
        i = 0,
        format = a[i++];

    function pad(str, len, chr, leftJustify)
    {
      var padding;

      if (!chr)
      {
        chr = ' ';
      }
      padding = (str.length >= len) ? '' : new Array(1 + len - str.length >>> 0).join(chr);
      return leftJustify ? str + padding : padding + str;
    }

    // justify()
    function justify(value, prefix, leftJustify, minWidth, zeroPad, customPadChar)
    {
      var diff = minWidth - value.length;
      if (diff > 0)
      {
        if (leftJustify || !zeroPad)
        {
          value = pad(value, minWidth, customPadChar, leftJustify);
        }
        else
        {
          value = value.slice(0, prefix.length) + pad('', diff, '0', true) + value.slice(prefix.length);
        }
      }
      return value;
    }

    // formatBaseX()
    function formatBaseX(value, base, prefix, leftJustify, minWidth, precision, zeroPad)
    {
      // Note: casts negative numbers to positive ones
      var number = value >>> 0;
      prefix = prefix && number && ({'2': '0b','8': '0', '16': '0x'}[base] || '');
      value = prefix + pad(number.toString(base), precision || 0, '0', false);
      return justify(value, prefix, leftJustify, minWidth, zeroPad);
    }

    // formatString()
    function formatString(value, leftJustify, minWidth, precision, zeroPad, customPadChar)
    {
      if (precision != null)
      {
        value = value.slice(0, precision);
      }
      return justify(value, '', leftJustify, minWidth, zeroPad, customPadChar);
    }

    // doFormat()
    function doFormat(substring, valueIndex, flags, minWidth, _, precision, type)
    {
      var number,
          prefix,
          method,
          textTransform,
          value,
          leftJustify = false,
          positivePrefix = '',
          zeroPad = false,
          prefixBaseX = false,
          customPadChar = ' ',
          flagsl = flags.length,
          j;

      if (substring == '%%')
      {
        return '%';
      }

      for (j = 0; flags && j < flagsl; j++)
      {
        switch (flags.charAt(j))
        {
          case ' ':
            positivePrefix = ' ';
            break;
          case '+':
            positivePrefix = '+';
            break;
          case '-':
            leftJustify = true;
            break;
          case "'":
            customPadChar = flags.charAt(j + 1);
            break;
          case '0':
            zeroPad = true;
            break;
          case '#':
            prefixBaseX = true;
            break;
        }
      }

      // parameters may be null, undefined, empty-string or real valued
      // we want to ignore null, undefined and empty-string values
      if (!minWidth)
      {
        minWidth = 0;
      }
      else if (minWidth == '*')
      {
        minWidth = +a[i++];
      }
      else if (minWidth.charAt(0) == '*')
      {
        minWidth = +a[minWidth.slice(1, -1)];
      }
      else
      {
        minWidth = +minWidth;
      }

      // Note: undocumented perl feature:
      if (minWidth < 0)
      {
        minWidth = -minWidth;
        leftJustify = true;
      }

      if (!isFinite(minWidth))
      {
        throw new Error('sprintf: (minimum-)width must be finite');
      }

      if (!precision)
      {
        precision = 'fFeE'.indexOf(type) > -1 ? 6 : (type == 'd') ? 0 : undefined;
      }
      else if (precision == '*')
      {
        precision = +a[i++];
      }
      else if (precision.charAt(0) == '*')
      {
        precision = +a[precision.slice(1, -1)];
      }
      else
      {
        precision = +precision;
      }

      // grab value using valueIndex if required?
      value = valueIndex ? a[valueIndex.slice(0, -1)] : a[i++];

      switch (type)
      {
        case 's':
          return formatString(String(value), leftJustify, minWidth, precision, zeroPad, customPadChar);
        case 'c':
          return formatString(String.fromCharCode(+value), leftJustify, minWidth, precision, zeroPad);
        case 'b':
          return formatBaseX(value, 2, prefixBaseX, leftJustify, minWidth, precision, zeroPad);
        case 'o':
          return formatBaseX(value, 8, prefixBaseX, leftJustify, minWidth, precision, zeroPad);
        case 'x':
          return formatBaseX(value, 16, prefixBaseX, leftJustify, minWidth, precision, zeroPad);
        case 'X':
          return formatBaseX(value, 16, prefixBaseX, leftJustify, minWidth, precision, zeroPad).toUpperCase();
        case 'u':
          return formatBaseX(value, 10, prefixBaseX, leftJustify, minWidth, precision, zeroPad);
        case 'i':
        case 'd':
          number = +value || 0;
          number = Math.round(number - number % 1); // Plain Math.round doesn't just truncate
          prefix = number < 0 ? '-' : positivePrefix;
          value = prefix + pad(String(Math.abs(number)), precision, '0', false);
          return justify(value, prefix, leftJustify, minWidth, zeroPad);
        case 'e':
        case 'E':
        case 'f': // Should handle locales (as per setlocale)
        case 'F':
        case 'g':
        case 'G':
          number = +value;
          prefix = number < 0 ? '-' : positivePrefix;
          method = ['toExponential', 'toFixed', 'toPrecision']['efg'.indexOf(type.toLowerCase())];
          textTransform = ['toString', 'toUpperCase']['eEfFgG'.indexOf(type) % 2];
          value = prefix + Math.abs(number)[method](precision);
          return justify(value, prefix, leftJustify, minWidth, zeroPad)[textTransform]();
        default:
          return substring;
      }
    }

    return format.replace(regex, doFormat);
  }

  return CangoCore;    // return the augmented Cango object, over-writing the existing

}(Cango));    // Take the existing Cango object and add Axes drawing methods
